/*----------------------------------------------------------------------------------------

	Introduction to TigerSHARC Multiprocessing Systems Using VisualDSP++
	Maikel K.B., DSP-Europe 
	June 2002

	ID0.C

	This program demonstrates some of the Multiprocessing features handled by VisualDSP++ Tools.
	It shows the use of TigerSHARC in an MP System by performing the following tasks: 

		*	Setting up of control registers: SYSCON and SDRCON
	
		*	ID Checking routine to ensure that the code has been loaded into the right DSP

		*	Go into endless loop and wait for Vector Interrupt (VIPRT) to be generated by DSP ID1

		*	Resume program execution after detecting and servicing Vector Interrupt generated by ID1.
		
		*	Copy the contents of tx_ID0 to rx_ID1 (internal memory of DSP ID1 via the MMS) and to 
			the	shared external memory (shared_data buffer declared in SDRAM) using DMA transfers.
 	
		*	Finally, a broadcast write to all DSP's internal memory is performed. For this, 
			the bus lock feature and DMA transfers are used.

		*	After completion, it goes back into an endless loop.

------------------------------------------------------------------------------------------*/

#define __NOUNDERSCORES__							// Depending on the compiler patch installed, the register names as defined in sysreg.h 
													// may now have two underscores prepended so that they do not restrict the namespace 
													// of user programs. If the old names are required as in this example(without the 
													// underscores prepended), 
													// define the preprocessor macro __NOUNDERSCORES__ before including sysreg.h.
													// Only required with compiler patch 6.1.7 or later

#include "sysreg.h"
#include "signal.h"
#include "stdio.h"

#define TAPS 16

extern int shared_data[TAPS*2];						// Data array defined in external shared memory (SDRAM)
extern int rx_ID1[TAPS];							// ID1 receive buffer (ID1.c) for external port (DMA) data transfer from ID0
extern int (*Ptr_to_Func)();						// Pointer to function declared in ID1 (ID1.c)

/* DM data section in Memory Block 1 */
section ("data1") int dm tx_ID0[TAPS] = {			// ID0 transmit buffer for external port (DMA) data transfers to ID1 
#include "tx_ID0.dat"								// and shared external memory (first half of buffer)
};

/* DM data section in Memory Block 2 */
section ("data2") int dm Broadcast_rx_ID0[TAPS];	// ID0 broadcast receive buffer for broadcast transfer from its own
section ("data2") int dm rx_ID0[TAPS];				// ID0 receive buffer for core (IALU) data transfer from ID1


struct TCB {
	int *DI;										// index
	int DX;											// count and stride in x direction
	int DY;											// count and stride in y direction
	int DP;											// DMA control word
};
struct TCB TCB_temp;								// Temp structure for programming TCBs

__builtin_quad q;									// Temp quad for programming TCBs

void Check_ID_Number();								// Prototype for Check ID Number routine
void Broadcast_write();								// Prototype for Broadcast write
void ID0_to_SDRAM_DMA();							// Prototype for ID0 to SDRAM DMA
void ID0_to_ID1_DMA();								// Prototype for ID0 to ID1 DMA
void broadcast_dma1_int();							// Prototype for Broadcast DMA1 interrupt service routine
void dma0_int();									// Prototype for DMA0 interrupt service routine
void dma1_int();									// Prototype for DMA1 interrupt service routine
int VIRPT_ISR_ID0();								// Prototype for VIRPT interrupt service routine

void main(void)
{
	int i=0,IMASKH_temp=0,IMASKL_temp=0;

	__builtin_sysreg_write(SYSCON,0x003a79e7);		// Initial value after reset (0x279e7) modified 
													// for 64-bit Bus width configuration
								
	__builtin_sysreg_write(SDRCON,0x00005913);		// Enable SDRAM: ENA=1, CAS=2 CL, pipedepth=0,			
										 			// page=512w, ref.rate=1200, trp=2, tras=5, init=1

	/* Enable DMA0, DMA1 and VIRPT interrupts */

	IMASKH_temp = __builtin_sysreg_read(IMASKH);	// Read the value of IMASKH
	IMASKH_temp = IMASKH_temp | 0x00010000;			// Enable VIRPT	
	__builtin_sysreg_write(IMASKH,IMASKH_temp);		// Write value back to IMASKH

	IMASKL_temp = __builtin_sysreg_read(IMASKL);	// Read the value of IMASKL
	IMASKL_temp = IMASKL_temp | 0x0000C000;			// Enable DMA0 & DMA1
	__builtin_sysreg_write(IMASKL,IMASKL_temp);		// Write value back to IMASKL

	Check_ID_Number();								// Check Code loaded into right DSP

	Ptr_to_Func = &VIRPT_ISR_ID0;					// Provide ISR address in ID0 to ID1 for VIRPT interrupt
	__builtin_idle();								// Wait until VIRPT interrupt is generated by ID1

	ID0_to_SDRAM_DMA();								// Perform DMA transfer from ID0 to SDRAM
	ID0_to_ID1_DMA();								// Perform DMA transfer from ID0 to ID1
	Broadcast_write();								// Perform Broadcast write to ALL DSPs (ID0 & ID1)

	while(1)										// Enter Endless loop
		i=i-i;	
}

/*********************************************************************************************************************************************/
/*	ID Checking Routine ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void Check_ID_Number()
{		

	int IDx=0;
	int SYSTAT_ID_bits=0x7;											// ID value in SYSTAT bits 2-0 -> 0111 = 0x7
	int i=0;
	
	IDx = (__builtin_sysreg_read(SYSTAT) & SYSTAT_ID_bits);			// read SYSTAT value and extract ID number 

	if (IDx != 0)													// is this DSP with ID0? 
	{	
		printf("\n ID0 code - Executable loaded into wrong DSP \n"); 
		while(1)
			i=i-i;
	}

	return;
}

/*********************************************************************************************************************************************/
/*	ID0 to SDRAM DMA0 transfer ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void ID0_to_SDRAM_DMA()
{		
	interrupt(SIGDMA0, dma0_int);					// Assign isr to DMA channel 0
														
	TCB_temp.DI = tx_ID0;							// index points to source buffer in ID0
	TCB_temp.DX = 4 | (TAPS << 16);					// modify is 4 for quad-word transfers, count is TAPS and must be shifted to upper half 
	TCB_temp.DY = 0;								// only a 1 dimension DMA
	TCB_temp.DP = 0x47000000;						// control word set for quad-word transfers to internal memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCS0, q);				// program the TCBs

	TCB_temp.DI = shared_data;						// index points to first half of buffer in SDRAM
	TCB_temp.DP = 0x87000000;						// control word set for quad-word transfers to external memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCD0, q);				// program the TCBs

	return;
}

/*********************************************************************************************************************************************/
/*	ID0 to ID1 DMA1 Transfer ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void ID0_to_ID1_DMA()
{		
	interrupt(SIGDMA1, dma1_int);					// Assign isr to DMA channel 1
	
	TCB_temp.DI = tx_ID0;							// index points to source buffer in ID0
	TCB_temp.DX = 4 | (TAPS << 16);					// modify is 4 for quad-word transfers, count is TAPS and must be shifted to upper half 
	TCB_temp.DY = 0;								// only a 1 dimension DMA
	TCB_temp.DP = 0x47000000;						// control word set for quad-word transfers to internal memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCS1, q);				// program the TCBs

	TCB_temp.DI = rx_ID1;							// index points to destination buffer in ID1 
	TCB_temp.DP = 0x87000000;						// control word set for quad-word transfers to external memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCD1, q);				// program the TCBs
	
	return;
}

/*********************************************************************************************************************************************/
/*	Broadcast write to ALL DSPs (ID0 & ID1) ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void Broadcast_write()	
{	
	int MMS_Broadcast = 0x1C00000;					// Broadcast MMS Offset

	__builtin_sysreg_write(BUSLK,0x1);				// Lock bus before performing broadcast write in order to prevent 
												 	// external devices from accessing data while being modified
	interrupt(SIGDMA1, broadcast_dma1_int);			// Assign isr to DMA channel 1

	TCB_temp.DI = tx_ID0;							// index points to source buffer in ID0
	TCB_temp.DX = 4 | (TAPS << 16);					// modify is 4 for quad-word transfers, count is TAPS and must be shifted to upper half 
	TCB_temp.DY = 0;								// only a 1 dimension DMA
	TCB_temp.DP = 0x47000000;						// control word set for quad-word transfers to internal memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCS1, q);				// program the TCBs

	TCB_temp.DI = Broadcast_rx_ID0+MMS_Broadcast;	// index points to destination buffer in ID0
	TCB_temp.DP = 0x87000000;						// control word set for quad-word transfers to external memory with interrupt enabled

	q = __builtin_compose_128((long long)TCB_temp.DI | (long long)TCB_temp.DX << 32, (long long)(TCB_temp.DY | (long long)TCB_temp.DP << 32));
	__builtin_sysreg_write4(DCD1, q);				// program the TCBs
	
	return;
}

/*********************************************************************************************************************************************/
/*	DMA0 Interrupt Service Routine ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void dma0_int()										
{		
	printf("\nExternal memory DMA from ID0 to ID1 completed\n");	
	return;
}

/*********************************************************************************************************************************************/
/*	DMA1 Interrupt Service Routine ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void dma1_int()										
{		
	printf("\nExternal memory DMA from ID0 to SDRAM completed\n"); 
	return;
}

/*********************************************************************************************************************************************/
/*	Broadcast DMA 1 Interrupt Service Routine ***********************************************************************************************************/
/*********************************************************************************************************************************************/
void broadcast_dma1_int()										
{		
	
	__builtin_sysreg_write(BUSLK,0x0);
	printf("\nBroadcast write from ID0 to ALL DSPs completed\n");	
	return;
}

/*********************************************************************************************************************************************/
/*	VIRPT Interrupt Service Routine ***********************************************************************************************************/
/*********************************************************************************************************************************************/
int VIRPT_ISR_ID0()										
{
			
	printf("\nVector Interrupt generated by ID1, serviced by ID0 sucessfully\n"); 
	ISR_RETURN();
	return(0);
}

